@@ -4,7 +4,7 @@ import threading
 from abc import ABC, abstractmethod
 from functools import lru_cache, partial
 from inspect import signature
-from typing import TYPE_CHECKING, Any, Callable, Type, cast
+from typing import TYPE_CHECKING, Any, Callable, Type, cast, get_type_hints
 
 import anyio
 
@@ -222,6 +222,7 @@ class BaseEvent:
 
     def __str__(self):
         str_list = []
+        slot: Any
         for slot in self.__slots__:
             val = str(getattr(self, slot))
             str_list.append(f"{slot}: {val}")
@@ -252,3 +253,48 @@ def process_event(value: Any, doc: Doc) -> Any:
 def count_parameters(func: Callable) -> int:
     """Count the number of parameters in a callable"""
     return len(signature(func).parameters)
+
+
+class Typed:
+    _: Any
+
+    def __init__(self) -> None:
+        self.__dict__["annotations"] = {
+            name: _type
+            for name, _type in get_type_hints(type(self).mro()[0]).items()
+            if name != "_"
+        }
+
+    if not TYPE_CHECKING:
+
+        def __getattr__(self, key: str) -> Any:
+            annotations = self.__dict__["annotations"]
+            if key not in annotations:
+                raise AttributeError(f'"{type(self).mro()[0]}" has no attribute "{key}"')
+            expected_type = annotations[key]
+            if hasattr(expected_type, "mro") and Typed in expected_type.mro():
+                return expected_type(self._[key])
+            return self._[key]
+
+        def __setattr__(self, key: str, value: Any) -> None:
+            if key == "_":
+                self.__dict__["_"] = value
+                return
+            annotations = self.__dict__["annotations"]
+            if key not in annotations:
+                raise AttributeError(f'"{type(self).mro()[0]}" has no attribute "{key}"')
+            expected_type = annotations[key]
+            if hasattr(expected_type, "__origin__"):
+                expected_type = expected_type.__origin__
+            if hasattr(expected_type, "__args__"):
+                expected_types = expected_type.__args__
+            else:
+                expected_types = (expected_type,)
+            if type(value) not in expected_types:
+                raise TypeError(
+                    f'Incompatible types in assignment (expression has type "{expected_type}", '
+                    f'variable has type "{type(value)}")'
+                )
+            if isinstance(value, Typed):
+                value = value._
+            self._[key] = value
